from gettext import gettext as _

from fastapi import status


class CustomI18NException(Exception):
    message: str
    error_code: int

    def __init__(
        self,
        message: str | None = None,
        err_code: int | None = None,
        *args,
        **kwargs,
    ):
        if message:
            self.message = message
        if err_code:
            self.error_code = err_code

    def __str__(self):
        return f"{self.__class__.__name__}({self.message})"

    @property
    def formatted_message(self):
        return self.message.format(**self.__dict__)


# Global
class SettingNotSet(CustomI18NException):
    message = _("{name} not set")
    error_code = status.HTTP_503_SERVICE_UNAVAILABLE
    name: str

    def __init__(self, name: str):
        self.name = name
        super().__init__(self.message, self.error_code)


class DoesNotExist(CustomI18NException):
    message = _("{name} does not exist")
    error_code = status.HTTP_404_NOT_FOUND
    name: str

    def __init__(self, name: str):
        self.name = name
        super().__init__(self.message, self.error_code)


class OnlyAdminOrOwnerAllowed(CustomI18NException):
    message = _("Only admin or owner allowed to access")
    error_code = status.HTTP_403_FORBIDDEN


class OnlyOwnerAllowed(CustomI18NException):
    message = _("Only owner of {resource} allowed to {action}")
    error_code = status.HTTP_403_FORBIDDEN
    resource: str
    action: str

    def __init__(self, resource: str, action: str):
        self.resource = resource
        self.action = action
        super().__init__(self.message, self.error_code)



from pydantic import BaseModel, Field, computed_field, field_validator

ADMIN_ROLE = "ADMIN"


class Role(BaseModel):
    name: str
    users: list[str] | None = None


class UserClaims(BaseModel):
    tenant: str = Field(alias="owner")
    id: str
    username: str = Field(..., validation_alias="name")
    email: str | None = None
    phone: str | None = None
    roles: list[Role] = Field(default_factory=list)

    @field_validator("roles", mode="before")
    @classmethod
    def validate_roles(cls, v: list[Role] | None) -> list[Role]:
        if not v:
            return []
        return v

    @computed_field  # type: ignore[misc]
    @property
    def is_admin(self) -> bool:
        return any(role.name == ADMIN_ROLE for role in self.roles or [])



from collections.abc import Callable, Coroutine
from typing import Annotated, Any

import httpx
from fastapi import Depends, HTTPException, Request
from fastapi.security import OAuth2PasswordBearer

from fastloom.auth.depends import JWTAuth, OptionalJWTAuth
from fastloom.auth.introspect.schema import IntrospectionResponse
from fastloom.auth.protocols import SidecarSettings
from fastloom.auth.schemas import UserClaims


class OptionalVerifiedAuth(OptionalJWTAuth):
    settings: SidecarSettings
    _oauth2_schema: OAuth2PasswordBearer | None = None

    def __init__(self, settings: SidecarSettings):
        super().__init__(settings)

    async def _introspect(
        self, token: Annotated[str, Depends(_oauth2_schema)]
    ):
        async with httpx.AsyncClient() as client:
            response: httpx.Response = await client.post(
                f"{self.settings.IAM_SIDECAR_URL}/introspect",
                json=dict(token=token),
            )
        if response.status_code != 200:
            raise HTTPException(status_code=403, detail=response.text)
        data = IntrospectionResponse.model_validate(response.json())
        if not data.active:
            raise HTTPException(status_code=403, detail="Inactive token")

    async def _acl(
        self, request: Request, token: Annotated[str, Depends(_oauth2_schema)]
    ) -> None:
        async with httpx.AsyncClient() as client:
            response: httpx.Response = await client.post(
                url=f"{self.settings.IAM_SIDECAR_URL}/acl",
                json={
                    "token": token,
                    "endpoint": request.url.path,
                    "method": request.method,
                },
            )

        if response.status_code != 200:
            raise HTTPException(status_code=403, detail=response.text)
        if not response.json():
            raise HTTPException(status_code=403)

    @property
    def get_claims(
        self,
    ) -> Callable[..., Coroutine[Any, Any, UserClaims | None]]:
        async def _inner(
            request: Request, token: str | None = Depends(self._oauth2_schema)
        ) -> UserClaims | None:
            if token is None:
                return None
            await self._introspect(token)
            await self._acl(request, token)
            return await super(OptionalVerifiedAuth, self).get_claims(token)

        return _inner


class VerifiedAuth(JWTAuth, OptionalVerifiedAuth):
    @property
    def get_claims(self) -> Callable[..., Coroutine[Any, Any, UserClaims]]:
        async def _inner(
            request: Request, token: str = Depends(self._oauth2_schema)
        ) -> UserClaims:
            await self._introspect(token)
            await self._acl(request, token)
            return await super(VerifiedAuth, self).get_claims(token)

        return _inner



import pkgutil
from collections.abc import Sequence
from importlib import import_module
from itertools import chain
from types import ModuleType
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from beanie import Document, UnionDoc, View
else:
    try:
        from beanie import Document, UnionDoc, View
    except ImportError:
        from pydantic import (
            BaseModel as Document,
        )
        from pydantic import (
            BaseModel as UnionDoc,
        )
        from pydantic import (
            BaseModel as View,
        )


async def get_mongo_client(mongo_uri: str):
    from pymongo import AsyncMongoClient

    return AsyncMongoClient(
        mongo_uri,
        tz_aware=True,
        connectTimeoutMS=1000,
        serverSelectionTimeoutMS=5000,
    )


def get_models(
    module: ModuleType,
) -> list[type[Document] | type[UnionDoc] | type[View]]:
    if (
        module.__spec__ is None
        or not module.__spec__.submodule_search_locations
    ):
        return [
            x
            for x in vars(module).values()
            if isinstance(x, type)
            and issubclass(x, Document | UnionDoc | View)
            and x not in [Document, UnionDoc, View]
        ]

    return list(
        chain.from_iterable(
            get_models(import_module(f"{module.__name__}.{i.name}"))
            for i in pkgutil.iter_modules(module.__path__)
        )
    )


async def init_db(
    database_name: str,
    models: Sequence[type[Document] | type[UnionDoc] | type[View] | str],
    mongo_uri: str,
):
    from beanie import init_beanie

    client = await get_mongo_client(mongo_uri)
    db = client[database_name]
    await init_beanie(db, document_models=models, recreate_views=True)


async def destroy_db(
    database_name: str,
    models: list[Document],
    mongo_uri: str,
    drop_database: bool = False,
):
    client = await get_mongo_client(mongo_uri)
    db = client[database_name]
    if not drop_database:
        for model in models[1:]:  # Skip pre-populated Province collection
            await db.drop_collection(model.Settings.name)
    else:
        await client.drop_database(database_name)



import importlib
import importlib.util
import logging
import os
import re
import signal
import sys
from functools import lru_cache
from pathlib import Path
from re import Pattern
from typing import TYPE_CHECKING

from pydantic import BaseModel

if TYPE_CHECKING:
    from fastloom.launcher.schemas import App

SettingsCls = type[BaseModel]


def _dynamic_import(name: str):
    if (cwd := Path.cwd()) not in sys.path:
        sys.path.insert(0, str(cwd))
    module_name = name
    module_path = Path.cwd() / f"{name}.py"

    spec = importlib.util.spec_from_file_location(module_name, module_path)
    if spec is None:
        raise ImportError(f"Failed to import {module_name}, {module_path}")
    app_module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = app_module
    assert spec.loader is not None
    spec.loader.exec_module(app_module)
    return app_module


def get_app() -> "App":
    return _dynamic_import("app").app


@lru_cache
def _get_settings_module():
    return _dynamic_import("settings")


def get_settings_cls() -> SettingsCls:
    return _get_settings_module().Settings


def get_tenant_cls() -> SettingsCls:
    try:
        return _get_settings_module().TenantSettings
    except AttributeError:
        logging.debug("No TenantSettings Found In settings.py")
        return BaseModel


class EndpointFilter(logging.Filter):
    def __init__(
        self, excluded_endpoints: tuple[Pattern | str, ...] = ()
    ) -> None:
        super().__init__()
        self.excluded_endpoints = excluded_endpoints

    def filter(self, record: logging.LogRecord) -> bool:
        if not record.args or len(record.args) < 3:
            return True
        endpoint: str = record.args[2]  # type: ignore[assignment, index]
        return not any(
            re.match(exp, endpoint) for exp in self.excluded_endpoints
        )


def reload_app():
    import inspect
    from pathlib import Path

    from fastloom.launcher.settings import LauncherSettings
    from fastloom.tenant.settings import ConfigAlias as Configs

    Path(
        next(
            filter(
                lambda x: __name__.split(".")[0] not in x.filename,
                inspect.stack(),
            )
        ).filename
    ).touch()
    if not Configs[LauncherSettings].general.DEBUG:
        os.kill(os.getppid(), signal.SIGHUP)



import logging
import os
from enum import Enum
from typing import Any, Generic, Literal, TypeVar
from uuid import UUID

from beanie import Insert, Replace, SaveChanges, Update, after_event
from pydantic import BaseModel, PrivateAttr, create_model, model_validator

from fastloom.db.schemas import BaseDocument
from fastloom.signals.depends import RabbitSubscriber

logger = logging.getLogger(__name__)


class Operations(str, Enum):
    UPDATE = "update"
    SAVE = "save"


T = TypeVar("T", bound=BaseDocument)


class SignalMessage(BaseModel, Generic[T]):
    instance: T
    changes: dict[str, Any]
    operation: Operations


class BaseDocumentSignal(BaseDocument):
    """
    Assumes that this mixin is used with `BaseDocument` subclasses and
    `BaseDocument` has full state management
    """

    _sent_events: set[tuple[UUID, Operations]] = PrivateAttr(
        default_factory=set
    )

    @model_validator(mode="after")
    def validate_state_management(self):
        self.check_state_management()
        return self

    async def _publish(self, message: SignalMessage):
        if self.revision_id is None:
            return
        _event_key = (
            self.revision_id,  # type: ignore[attr-defined]
            message.operation,
        )
        if _event_key in self._sent_events:
            logger.debug(f"prevented publishing event: {_event_key}")
            return
        logger.debug(f"publishing event: {_event_key}")
        await self.get_publisher(message.operation).publish(message)
        self._sent_events.add(_event_key)

    @classmethod
    def get_subscription_topic(cls, operation: Operations):
        project_name = os.getenv("PROJECT_NAME")
        if not project_name:
            raise ValueError("PROJECT_NAME environment variable is not set")
        return f"{project_name}.{cls.get_collection_name()}.{operation.value}"  # type: ignore[attr-defined]  # noqa

    @classmethod
    def check_state_management(cls):
        if not (
            cls.use_state_management() and cls.state_management_save_previous()
        ):
            raise ValueError(
                f"State management is not enabled for {cls.__name__}"
            )

    @classmethod
    def get_publisher(cls, operation: Operations):
        class_changes = create_model(  # type: ignore[call-overload]
            f"{cls.__name__}Changes",
            __base__=cls,
            field_definitions={
                field_name: (f"{field_info.annotation | None}", None)
                for field_name, field_info in cls.model_fields.items()
            },
        )
        class_instance = create_model(  # type: ignore[call-overload]
            f"{cls.__name__}Instance",
            __base__=cls,
            field_definitions={
                field_name: (field_info.annotation, ...)
                for field_name, field_info in cls.model_fields.items()
            },
        )
        # NOTE: ^ to avoid discriminator error because AsyncAPI has issues with
        # it. Otherwise we would've defined __base__ of
        # `narrowed_signal_message` to `SignalMessage[cls]` and wouldn't have
        # overridden `instance`.
        narrowed_signal_message = create_model(
            f"{cls.__name__}SignalMessage",
            __base__=SignalMessage,
            instance=(class_instance, ...),
            changes=(class_changes, ...),
            operation=(Literal[operation], ...),
        )
        return RabbitSubscriber.publisher(
            routing_key=cls.get_subscription_topic(operation),
            schema=narrowed_signal_message,
        )


class SignalsSave(BaseDocumentSignal):
    @after_event(Insert)
    async def _publish_post_save(cls):
        await cls.publish_post_save()

    async def publish_post_save(self):
        await self._publish(
            SignalMessage(
                instance=self,
                changes=self.get_previous_changes(),
                operation=Operations.SAVE,
            ),
        )


class SignalsUpdate(BaseDocumentSignal):
    @after_event(Replace, SaveChanges, Update)
    async def _publish_post_update(self):
        await self.publish_post_update()

    async def publish_post_update(self):
        await self._publish(
            SignalMessage(
                instance=self,
                changes=self.get_previous_changes(),
                operation=Operations.UPDATE,
            )
        )
        # TODO: We can maybe send separate signals for each field change?



import gettext
from collections.abc import Callable
from pathlib import Path

from babel import Locale
from babel.support import Translations
from fastapi.responses import JSONResponse
from jinja2 import Environment, FileSystemLoader
from starlette.requests import Request

from fastloom.i18n.base import CustomI18NException
from fastloom.i18n.types import Languages


def parse_locale(locale: str | None) -> str:
    try:
        return Locale.parse(locale).language
    except (ValueError, TypeError):
        return "en"


def set_locale(locale: str | None) -> Callable[[str], str]:
    language = parse_locale(locale)

    locale_dir: Path = Path.cwd() / Path("locale")
    trans = gettext.translation(
        "messages",
        localedir=locale_dir,
        languages=[language],
        fallback=True,
    )
    trans.install()
    return trans.gettext


async def i18n_exception_handler(
    request: Request, exc: CustomI18NException
) -> JSONResponse:
    locale: str | None = request.headers.get("Accept-Language")
    gt: Callable[[str], str] = set_locale(locale)
    return JSONResponse(
        status_code=exc.error_code,
        content={
            "detail": exc.message,
            "message": exc.message,
            "message_tr": gt(exc.formatted_message),
        },
    )


def get_template(
    template_name: str, locale: str | None, *args, **kwargs
) -> str:
    language = parse_locale(locale)
    env = Environment(
        loader=FileSystemLoader(Path.cwd() / "templates"),
        extensions=["jinja2.ext.i18n"],
    )
    template = env.get_template(template_name)
    translations = Translations.load("locale", [language])
    return template.render(
        gettext=translations.gettext, lang=language, **kwargs
    )


def lang_dict(fa: str, en: str | None = None) -> dict[Languages, str]:
    """Create a language dictionary using FA as base and DEFAULT"""
    result = {Languages.BASE: fa, Languages.FA: fa}
    result[Languages.EN] = en if en is not None else fa
    return result



from collections.abc import Callable, Coroutine
from functools import partial
from typing import Any


class MongoConnectionError(Exception): ...


async def check_mongo_connection(mongo_uri: str) -> None:
    from pymongo import AsyncMongoClient

    try:
        client: AsyncMongoClient = AsyncMongoClient(mongo_uri, timeoutms=2000)
        await client.admin.command("ping")
    except Exception as er:
        raise MongoConnectionError(f"MongoDB connection error: {er}") from er


def get_healthcheck(
    mongo_uri: str,
) -> Callable[[], Coroutine[Any, Any, None]]:
    return partial(check_mongo_connection, mongo_uri=mongo_uri)



from collections.abc import Callable, Coroutine
from typing import Any

from fastapi import APIRouter, FastAPI
from fastapi.responses import JSONResponse


def init_healthcheck(
    app: FastAPI,
    healthcheck_handlers: list[Callable[[], Coroutine[Any, Any, None]]],
    prefix: str = "",
) -> None:
    router = APIRouter()

    @router.get(f"{prefix}/healthcheck")
    async def healthcheck_endpoint() -> JSONResponse:
        for handler in healthcheck_handlers:
            await handler()

        return JSONResponse(content={"status": "ok"})

    app.include_router(router, tags=["Healthcheck"])



from typing import Protocol

from pydantic import AnyHttpUrl


class TenantNameSchema(Protocol):
    name: str


class TenantHostSchema(TenantNameSchema, Protocol):
    website_url: AnyHttpUrl | list[AnyHttpUrl]


class TenantMonitoringSchema(Protocol):
    PROJECT_NAME: str



from pydantic import AnyHttpUrl

from fastloom.settings.base import MonitoringSettings


class ObservabilitySettings(MonitoringSettings):
    SENTRY_ENABLED: int = 0
    OTEL_ENABLED: int = 0
    SENTRY_DSN: AnyHttpUrl | None = None
    METRICS: bool = False



from os import getenv
from typing import Callable


def get_env_or_err(field_name: str) -> Callable[[], str]:
    def _inner() -> str:
        value = getenv(field_name)
        if value is None:
            raise ValueError(
                f"{field_name} must be set in environment or config"
            )
        return value

    return _inner



import re
from typing import Annotated

from pydantic import AfterValidator, Field

PHONE_REGEX = r"^(\+|00)\d{1,2}\s?((\(\d{3}\))|\d{3})[\s.-]?\d{3}[\s.-]?\d{4}$"
EMAIL_REGEX = r"^[a-zA-Z0-9._%+-]+@[a-zA-Z0-9.-]+\.[a-zA-Z]{2,}$"


class PhoneValidation:
    @classmethod
    def phone_validator(cls, string: str | None) -> str | None:
        if string and re.match(PHONE_REGEX, string):
            string = re.sub(r"^00", "+", string)
            string = re.sub(r"[()\s.-]", "", string)
            return string
        return None

    @classmethod
    def phone_validator_or_exc(cls, string: str) -> str:
        if phone := cls.phone_validator(string):
            return phone
        raise ValueError("Invalid phone number")


class EmailValidation:
    @classmethod
    def email_validator(cls, string: str) -> str | None:
        email_regex = EMAIL_REGEX
        if re.match(email_regex, string):
            return string
        return None


PhoneNumberOrNone = Annotated[
    str | None, AfterValidator(PhoneValidation.phone_validator)
]
ValidatedPhoneNumber = Annotated[
    str, AfterValidator(PhoneValidation.phone_validator_or_exc)
]


def _national_id_validator(nc: str) -> str | None:
    assert "12345678" not in nc
    assert len(nc) == 10
    assert nc.isdigit()
    assert nc not in (str(i) * 10 for i in range(10))
    rem = sum(int(c) * (10 - i) for i, c in enumerate(nc[:9])) % 11
    assert int(nc[9]) == ((11 - rem) if rem >= 2 else rem)
    return nc


NationalID = Annotated[str, AfterValidator(_national_id_validator)]

UserId = str

NoneField = Annotated[None, Field(None)]



import pkgutil
from collections.abc import Awaitable, Callable, Coroutine
from contextlib import asynccontextmanager
from importlib import import_module
from pathlib import Path
from types import ModuleType
from typing import TYPE_CHECKING, Any, Self, TypeVar

from fastloom.launcher.settings import LauncherSettings
from fastloom.tenant.handler import init_settings_endpoints

if TYPE_CHECKING:
    from beanie import View
    from beanie.odm.documents import Document
    from beanie.odm.union_doc import UnionDoc
else:
    try:
        from beanie import View
        from beanie.odm.documents import Document
        from beanie.odm.union_doc import UnionDoc
    except ImportError:
        from pydantic import BaseModel as Document
        from pydantic import BaseModel as UnionDoc
        from pydantic import BaseModel as View

from fastapi import APIRouter, FastAPI
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    computed_field,
    model_validator,
)
from starlette.requests import Request
from starlette.responses import Response
from starlette.types import Lifespan

from fastloom.db.healthcheck import get_healthcheck as db_hc
from fastloom.db.lifehooks import get_models, init_db
from fastloom.db.settings import MongoSettings
from fastloom.healthcheck.handler import init_healthcheck
from fastloom.i18n.base import CustomI18NException
from fastloom.i18n.handler import i18n_exception_handler
from fastloom.settings.base import FastAPISettings
from fastloom.signals.depends import RabbitSubscriber
from fastloom.signals.healthcheck import (
    get_healthcheck as signal_hc,
)
from fastloom.tenant.settings import ConfigAlias as Configs

E = TypeVar("E", bound=Exception)
Route = tuple[APIRouter, str, str]
SettingsCls = type[BaseModel]
Healthcheck = Callable[[], Coroutine[Any, Any, None]]
ExceptionHandler = Callable[[Request, E], Response | Awaitable[Response]]
ExceptionHandlerRegister = tuple[int | type[E], ExceptionHandler]


def default_lifespan():
    @asynccontextmanager
    async def _identity_gen(_):
        yield

    return _identity_gen


class App(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    signals_module: ModuleType | None = None
    models_module: ModuleType | None = None
    healthchecks: list[Healthcheck] = Field(default_factory=list)
    additional_instruments: list[Callable] = Field(default_factory=list)
    routes: list[Route] = Field(default_factory=list)
    models: list[type[Document] | type[UnionDoc] | type[View]] = Field(
        default_factory=list
    )
    lifespan_fn: Lifespan = Field(default_factory=default_lifespan)
    exception_handlers: list[ExceptionHandlerRegister] = Field(
        default_factory=list
    )

    @computed_field  # type: ignore[prop-decorator]
    @property
    def project_name(self) -> str | None:
        return self.models[0].__module__.split(".")[0] if self.models else None

    @property
    def root_router(self) -> APIRouter:
        _router = APIRouter()
        for router, prefix, name in self.routes:
            _router.include_router(router, prefix=prefix, tags=[name])
        return _router

    async def load(self):
        await self.load_db()
        self.load_signals()

    async def load_db(self):
        if not self.models:
            return
        await init_db(
            database_name=Configs[MongoSettings].general.MONGO_DATABASE,
            models=self.models + [Configs.tenant_schema.document],
            mongo_uri=Configs[MongoSettings].general.MONGO_URI,
        )

    def load_signals(self):
        if not self.signals_module:
            return
        if not hasattr(self.signals_module, "__path__"):
            return import_module(self.signals_module.__name__)
        for i in pkgutil.iter_modules(self.signals_module.__path__):
            target = f"{self.signals_module.__name__}.{i.name}"
            if i.ispkg:
                target_path = Path(self.signals_module.__path__[0]) / i.name
                for j in pkgutil.iter_modules([target_path]):
                    import_module(f"{target}.{j.name}")
            else:
                import_module(target)

    def load_healthchecks(self, app: FastAPI):
        handlers: list[Healthcheck] = [
            *self.healthchecks,
        ]

        if self.models:
            handlers.append(db_hc(Configs[MongoSettings].general.MONGO_URI))  # type: ignore[misc]
        if self.signals_module:
            handlers.append(signal_hc(RabbitSubscriber.router))

        init_healthcheck(app=app, healthcheck_handlers=handlers)
        # ^for docker and system healthcheck
        init_healthcheck(
            app=app,
            healthcheck_handlers=handlers,
            prefix=Configs[FastAPISettings].general.API_PREFIX,  # type: ignore[misc]
        )

    def load_system_endpoints(self, app: FastAPI):
        init_settings_endpoints(app=app, configs=Configs)
        if Configs[LauncherSettings].general.SETTINGS_PUBLIC:  # type: ignore[misc]
            init_settings_endpoints(
                app=app,
                configs=Configs,
                prefix=Configs[FastAPISettings].general.API_PREFIX,  # type: ignore[misc]
            )

    def load_exception_handlers(self, app: FastAPI):
        for exc_class_or_status_code, handler in (
            (CustomI18NException, i18n_exception_handler),
            *self.exception_handlers,
        ):
            app.exception_handler(exc_class_or_status_code)(handler)

    @model_validator(mode="after")
    def module_to_models(self) -> Self:
        if self.models_module is not None and not self.models:
            self.models = get_models(self.models_module)
        return self



import logging

import click


class ColoredFormatter(logging.Formatter):
    @staticmethod
    def get_colored_levelname(level_name, levelno):
        level_colors = {
            logging.DEBUG: lambda text: click.style(text, fg="bright_blue"),
            logging.INFO: lambda text: click.style(text, fg="green"),
            logging.WARNING: lambda text: click.style(text, fg="yellow"),
            logging.ERROR: lambda text: click.style(text, fg="red"),
            logging.CRITICAL: lambda text: click.style(text, fg="bright_red"),
        }
        return level_colors.get(levelno, lambda text: text)(level_name)

    def format(self, record):
        formatted_message = super().format(record)
        levelname = self.get_colored_levelname(
            record.levelname, record.levelno
        )
        name = click.style(record.name, fg="magenta")
        message = click.style(formatted_message, fg="bright_cyan")

        colorized_message = f"{levelname}: \t  {name} {message}"
        return colorized_message



from datetime import datetime
from typing import Annotated, Self

import bson
from beanie import (
    Document,
    Indexed,
    Insert,
    PydanticObjectId,
    Replace,
    SaveChanges,
    Update,
    before_event,
)
from fastapi import HTTPException, status
from pydantic import BaseModel, Field, computed_field, field_validator

from fastloom.date import utcnow


class CreatedAtSchema(BaseModel):
    created_at: datetime = Field(default_factory=utcnow)


class CreatedUpdatedAtSchema(CreatedAtSchema):
    """
    ONLY use this mixin in `beanie.Document` models since it uses
    @before_event decorator

    NOTE: `updated_at` doesn't get updated when `update_many` is called
    """

    updated_at: datetime | None = Field(default_factory=utcnow)
    # TODO ^ it shouldn't ideally be None, but some models used to save null
    # so first we have to make sure we cleared db from all such instances

    @before_event(Insert, Replace, SaveChanges, Update)
    async def update_updated_at(self):
        self.updated_at = utcnow()


class BaseDocument(Document):
    id: PydanticObjectId = Field(default_factory=bson.ObjectId, alias="_id")  # type: ignore[assignment] # noqa

    @classmethod
    async def get_or_404(cls, id: PydanticObjectId) -> Self:
        obj: Self | None = await cls.get(id)
        if obj is None:
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND,
                detail=f"{cls.__name__} not found",
            )
        return obj

    @classmethod
    async def find_one_or_404(cls, *args, **kwargs) -> Self:
        obj: Self | None = await cls.find_one(*args, **kwargs)
        if obj is None:
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND,
                detail=f"{cls.__name__} not found",
            )
        return obj


class BasePaginationQuery(BaseModel):
    offset: int | None = Field(None, ge=0)
    limit: int | None = Field(None, ge=0)

    @field_validator("limit", mode="after")
    @classmethod
    def convert_zero_limit(cls, v: int | None) -> int | None:
        return v or None

    @computed_field  # type: ignore[misc]
    @property
    def skip(self) -> int | None:
        if self.limit and self.offset is not None:
            return self.limit * self.offset
        return None


class PaginatedResponse[T](BaseModel):
    data: list[T] = Field(default_factory=list)
    count: int = Field(default=0, ge=0)


class BaseTenantSettingsDocument(Document, CreatedUpdatedAtSchema):
    id: Annotated[str, Indexed()]  # type: ignore[assignment]

    class Settings:
        name = "settings"



from aredis_om import Field, JsonModel

from fastloom.cache.lifehooks import RedisHandler

RedisHandler()


class BaseCache(JsonModel):
    class Meta:
        global_key_prefix = "cache"
        database = RedisHandler.redis
        model_key_prefix = "base"
        # ^should be overriden in sub

    @property
    async def invalidate(self):
        return await self.expire(0)


class BaseTenantSettingCache(BaseCache):
    id: str = Field(primary_key=True)


class HostTenantMapping(BaseCache, index=True):  # type: ignore[call-arg]
    host: str = Field(primary_key=True)
    tenant: str = Field(index=True)

    class Meta:
        model_key_prefix = "host_mapping"



from typing import Any, Self, cast

from pydantic import BaseModel, create_model
from pydantic.fields import FieldInfo


class SelfSustainingMeta(type):
    def __new__(mcls, name, bases, namespace):
        namespace["self"] = cast(object, None)  # reserve the slot
        return super().__new__(mcls, name, bases, namespace)

    def __getattr__(cls, name):
        if cls.self is None:
            raise AttributeError(f"{cls.__name__}.self is not initialized")

        return getattr(cls.self, name)

    def __setattr__(cls, name, value):
        if name == "self" or name in cls.__dict__:
            return super().__setattr__(name, value)
        if name == "__parameters__":
            return super().__setattr__(name, value)
        if cls.self is None:
            raise AttributeError(f"{cls.__name__}.self is not initialized")
        return setattr(cls.self, name, value)


class SelfSustaining(metaclass=SelfSustainingMeta):
    self: Self

    def __init__(self, *args, **kwargs):
        type(self).self = self  # store the singleton


def optional_fieldinfo(
    field: FieldInfo, strip: bool = False
) -> tuple[Any, FieldInfo]:
    field = field._copy()
    if field.is_required() or strip:
        field.default = None
        field.default_factory = None
        field.validate_default = False
        if field.annotation is not None:
            field.annotation = cast(type[Any], field.annotation | None)
    return field.annotation, field


def create_optional_model(
    model: type[BaseModel], strip: bool = False, name: str | None = None
) -> type[BaseModel]:
    return create_model(
        f"Optional{model.__name__}" if name is None else name,
        **{
            k: cast(Any, optional_fieldinfo(v, strip=strip))
            for k, v in model.model_fields.items()
        },
    )



from typing import Any, Generic, TypeVar

from pydantic import BaseModel, create_model
from pydantic.fields import FieldInfo

from fastloom.meta import create_optional_model, optional_fieldinfo

V = TypeVar("V", bound=BaseModel)
U = TypeVar("U", bound=BaseModel)
Z = TypeVar("Z", bound=BaseModel)


# [settings class , document class, cache class]
class SettingCacheSchema(Generic[V, U, Z]):
    model: type[V]
    config: type[BaseModel]
    optional: type[BaseModel]
    document: type[U]
    cache: type[Z]
    config_default: dict[str, Any] = {}

    def __init__(
        self, model: type[V], document_cls: type[U], cache_class: type[Z]
    ):
        self.model = model
        self.optional = create_optional_model(
            model, name=f"Optional{model.__name__}", strip=True
        )
        self.config = create_optional_model(
            model, name=f"OptionalConfig{model.__name__}"
        )
        self.document = create_model(
            f"{model.__name__}Document",
            __base__=(  # type: ignore[arg-type]
                self.optional,
                document_cls,
            ),
        )
        self.cache = create_model(
            f"{model.__name__}Cache",
            __base__=(  # type: ignore[arg-type]
                self.optional,
                cache_class,
            ),
            __cls_kwargs__={"index": True},
        )

    def validate(self, fetched: V) -> V:
        return self.model.model_validate(
            self.config_default | (fetched.model_dump(exclude_defaults=True))
        )

    def strip_defaults(self, fetched: V) -> dict[str, Any]:
        stripped = fetched.model_dump(exclude_defaults=True)
        for key in self.config_default:
            if key in stripped and stripped[key] == self.config_default[key]:
                del stripped[key]

        return stripped

    def get_schema(self) -> dict[str, Any]:
        fields: dict[str, FieldInfo] = {
            k: optional_fieldinfo(v, strip=True)[1]
            if k in self.config_default
            else v._copy()
            for k, v in self.model.model_fields.items()
        }
        schema_model: BaseModel = create_model(  # type: ignore[call-overload]
            f"{self.model.__name__}Schema",
            **{k: (v.annotation, v) for k, v in fields.items()},
        )
        return schema_model.model_json_schema()



from typing import Annotated

from pydantic import AnyHttpUrl, BaseModel, Field

WebsiteUrlType = AnyHttpUrl


class TenantMixin(BaseModel):
    tenant: str


class BaseTenantWithHostSettings(BaseModel):
    website_url: (
        WebsiteUrlType | Annotated[list[WebsiteUrlType], Field(min_length=1)]
    )



from collections.abc import Callable, Coroutine
from typing import Any, get_args

from beanie.operators import Set

from fastloom.file.models import FileReference
from fastloom.file.schema import FileMessage
from fastloom.signals.depends import RabbitSubscriber


def init_file_signals(
    file_ref_cls: type[FileReference],
) -> Callable[..., Coroutine[Any, Any, None]]:
    usages = get_args(file_ref_cls.model_fields["usage"].annotation)

    @RabbitSubscriber.multi_subscriber(
        routing_keys=[f"file.{usage}.created" for usage in usages],
        retry_backoff=True,
    )
    async def file_uploaded(files: list[FileMessage]):
        for file in files:
            new_file_ref = file_ref_cls(**file.model_dump(), matched=True)
            await file_ref_cls.find_one(
                file_ref_cls.name == file.name,
                file_ref_cls.usage == file.usage,
                file_ref_cls.tenant == file.tenant,
            ).upsert(
                Set(new_file_ref.model_dump(exclude={"id"})),
                on_insert=new_file_ref,
            )

    return file_uploaded



from enum import StrEnum
from typing import Annotated, Any

from pydantic import Field, PlainSerializer


def exclude_none_serializer(value: dict[Any, str | None]) -> dict[Any, str]:
    return {k: v for k, v in value.items() if v is not None}


class Languages(StrEnum):
    BASE = "BASE"
    EN = "EN"
    FA = "FA"


TranslatedValue = Annotated[str | None, Field(None, max_length=2000)]


Translation = Annotated[
    dict[Languages, TranslatedValue],
    Field(default_factory=dict),
    PlainSerializer(exclude_none_serializer, return_type=dict[Languages, str]),
]



from beanie import PydanticObjectId

from fastloom.file.models import FileReference
from fastloom.file.schema import FileIn


async def get_or_create_reference(
    file_ref_cls: type[FileReference], file_in: FileIn, tenant: str
) -> FileReference | None:
    if file_in is None:
        return None
    return await file_ref_cls.find_one(
        file_ref_cls.name == file_in.name, file_ref_cls.tenant == tenant
    ) or file_ref_cls(  # type: ignore[call-arg]
        id=PydanticObjectId(),
        name=file_in.name,
        matched=False,
        tenant=tenant,  # TODO: wtf is wrong with providing tenant? mypy blames
    )



import random
import typing
from pprint import pformat

from deepdiff import DeepDiff  # type: ignore[import-untyped]

if typing.TYPE_CHECKING:

    class ResponseType(typing.Protocol):
        status_code: int
        text: str
        is_success: bool


def assert_deep_diff(actual, expected, **options):
    try:
        import bson
        from beanie import PydanticObjectId

        options |= dict(
            ignore_type_in_groups=[(bson.ObjectId, PydanticObjectId, str)]
        )
    except ImportError:
        ...
    diff = DeepDiff(
        expected,
        actual,
        ignore_order=True,
        exclude_types={type(...)},  # ignore comparison with ellipsis
        verbose_level=2,
        ignore_type_subclasses=False,
        ignore_nan_inequality=True,
        **options,
    )
    assert not diff, pformat(diff)


def random_mobile_number():
    return f"0912{random.randint(1000000, 9999999)}"


def status_check(response: "ResponseType", status: int):
    assert response.status_code == status, (
        response.status_code,
        response.text,
    )


def assert_success(response: "ResponseType"):
    assert response.is_success, response.text



from re import Pattern

from pydantic import BaseModel


class LauncherSettings(BaseModel):
    APP_PORT: int = 8000
    DEBUG: bool = True
    WORKERS: int = 4
    SETTINGS_PUBLIC: bool = False
    LOGGING_EXCLUDED_ENDPOINTS: tuple[Pattern | str, ...] = (
        r"/api/\w+/healthcheck",
    )



from pathlib import Path
from typing import Protocol

from pydantic import AnyHttpUrl


class OAuth2Settings(Protocol):
    IAM_TOKEN_URL: AnyHttpUrl | Path


class SidecarSettings(OAuth2Settings, Protocol):
    IAM_SIDECAR_URL: AnyHttpUrl



import logging
from collections.abc import AsyncGenerator
from contextlib import asynccontextmanager
from logging import Logger

import uvicorn
from fastapi import FastAPI
from starlette.middleware.cors import CORSMiddleware

from fastloom.db.settings import MongoSettings
from fastloom.launcher.settings import LauncherSettings
from fastloom.launcher.utils import (
    EndpointFilter,
    get_app,
    get_settings_cls,
    get_tenant_cls,
)
from fastloom.monitoring import InitMonitoring, Instruments
from fastloom.observability.settings import ObservabilitySettings
from fastloom.settings.base import FastAPISettings
from fastloom.signals.depends import RabbitSubscriber, RabbitSubscriptable
from fastloom.signals.settings import RabbitmqSettings
from fastloom.tenant.settings import ConfigAlias as Configs

logger: Logger = logging.getLogger(__name__)


@asynccontextmanager
async def lifespan(app: FastAPI) -> AsyncGenerator[None, None]:
    service_app = get_app()
    await service_app.load()
    if Configs.cache_enabled:
        from aredis_om import Migrator

        await Migrator().run()
    async with service_app.lifespan_fn(app):
        yield


def initial_app():
    Configs(get_settings_cls(), get_tenant_cls())
    logging.getLogger("uvicorn.access").addFilter(
        EndpointFilter(
            Configs[LauncherSettings].general.LOGGING_EXCLUDED_ENDPOINTS
        )
    )
    if isinstance(Configs[RabbitSubscriptable].general, RabbitmqSettings):
        RabbitSubscriber(Configs[RabbitSubscriptable].general)
    else:
        logging.warning("Settings Does Not Inherit from RabbitmqSettings")
    # ^IMPORTANT:rabbit has to init first
    service_app = get_app()
    instruments = [Instruments.HTTPX, Instruments.REDIS]
    if isinstance(Configs.general, RabbitmqSettings):
        instruments.append(Instruments.RABBIT)
    if isinstance(Configs.general, MongoSettings):
        instruments.append(Instruments.MONGODB)
    if Configs[ObservabilitySettings].general.METRICS:
        instruments.append(Instruments.METRICS)
    with InitMonitoring(
        Configs[ObservabilitySettings].general,
        instruments=instruments + service_app.additional_instruments,
    ) as monitor:
        app = FastAPI(
            lifespan=lifespan,
            title=Configs[FastAPISettings].general.PROJECT_NAME,
            docs_url=f"{Configs[FastAPISettings].general.API_PREFIX}/docs",
            openapi_url=f"{Configs[FastAPISettings].general.API_PREFIX}/openapi.json",
        )
        monitor.instrument(app)

    app.add_middleware(
        CORSMiddleware,
        allow_origins=["*"],
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )

    service_app.load_exception_handlers(app)
    service_app.load_healthchecks(app)
    service_app.load_system_endpoints(app)

    app.include_router(
        service_app.root_router,
        prefix=Configs[FastAPISettings].general.API_PREFIX,
    )
    if isinstance(Configs[RabbitSubscriptable].general, RabbitmqSettings):
        app.include_router(RabbitSubscriber.router)
    return app


app = initial_app()


def main():
    uvicorn.run(
        app=f"{__name__}:app",
        host="0.0.0.0",
        port=Configs[LauncherSettings].general.APP_PORT,
        reload=Configs[LauncherSettings].general.DEBUG,
        workers=Configs[LauncherSettings].general.WORKERS,
    )



from contextlib import suppress
from os import getenv

from aredis_om import get_redis_connection
from redis import Redis
from redis.exceptions import ConnectionError

from fastloom.cache.settings import RedisSettings
from fastloom.meta import SelfSustaining


class RedisHandler(SelfSustaining):
    enabled: bool
    redis: Redis
    sync_redis: Redis

    def __init__(self):
        super().__init__()
        self.enabled = False
        settings = RedisSettings.model_validate(
            dict(redis_url=getenv("REDIS_OM_URL"))
            if getenv("REDIS_OM_URL")
            else {}
        )
        self.redis = get_redis_connection(url=str(settings.redis_url))
        self.sync_redis = Redis.from_url(url=str(settings.redis_url))
        with suppress(ConnectionError):
            self.enabled = self.sync_redis.ping()



from pathlib import Path

from pydantic import (
    AnyHttpUrl,
    BaseModel,
    Field,
    computed_field,
)

from fastloom.settings.utils import get_env_or_err


class ProjectSettings(BaseModel):
    PROJECT_NAME: str = Field(
        default_factory=get_env_or_err("PROJECT_NAME"),
    )


class FastAPISettings(ProjectSettings):
    DEBUG: bool = True

    @computed_field  # type: ignore[misc]
    @property
    def API_PREFIX(self) -> str:
        return f"/api/{self.PROJECT_NAME}"


class IAMSettings(BaseModel):
    IAM_SIDECAR_URL: AnyHttpUrl = Field(
        AnyHttpUrl("http://iam:8000/api/iam/sidecar")
    )
    IAM_TOKEN_URL: AnyHttpUrl | Path = Path("/api/iam/auth/login/basic")


class MonitoringSettings(ProjectSettings):
    ENVIRONMENT: str


class BaseGeneralSettings(IAMSettings, MonitoringSettings): ...



from typing import TYPE_CHECKING, Any

from aio_pika import IncomingMessage
from faststream import BaseMiddleware
from faststream.broker.message import StreamMessage
from faststream.opentelemetry.middleware import (
    BaseTelemetryMiddleware,
    TelemetryMiddleware,
)
from faststream.rabbit.opentelemetry.provider import (
    RabbitTelemetrySettingsProvider,
)
from opentelemetry.metrics import Meter, MeterProvider
from opentelemetry.trace import TracerProvider
from pydantic import BaseModel

if TYPE_CHECKING:
    from faststream.types import AnyDict, AsyncFunc


def message_body_to_str(message_body) -> str:
    if isinstance(message_body, BaseModel):
        return message_body.model_dump_json()
    elif isinstance(message_body, str):
        return message_body
    else:
        return str(message_body)


class RabbitPayloadTelemetrySettingsProvider(RabbitTelemetrySettingsProvider):
    def get_publish_attrs_from_kwargs(
        self,
        kwargs: "AnyDict",
    ) -> "AnyDict":
        ret = super().get_publish_attrs_from_kwargs(kwargs)
        message_body = kwargs.pop("message_body")
        try:
            message_body_str = message_body_to_str(message_body)
        except ValueError:
            return ret
        return ret | {"messaging.message.body": message_body_str}

    def get_consume_attrs_from_message(
        self, msg: "StreamMessage[IncomingMessage]"
    ) -> "AnyDict":
        return super().get_consume_attrs_from_message(msg) | {
            "messaging.message.body": msg.body.decode()
        }


class PayloadTelemetryMiddleware(BaseTelemetryMiddleware):
    async def publish_scope(
        self, call_next: "AsyncFunc", msg: Any, *args: Any, **kwargs: Any
    ) -> Any:
        return await super().publish_scope(
            call_next, msg, *args, **(kwargs | dict(message_body=msg))
        )


class RabbitPayloadTelemetryMiddleware(TelemetryMiddleware):
    def __init__(
        self,
        *,
        tracer_provider: TracerProvider | None = None,
        meter_provider: MeterProvider | None = None,
        meter: Meter | None = None,
    ) -> None:
        super().__init__(
            settings_provider_factory=(
                lambda _: RabbitPayloadTelemetrySettingsProvider()
            ),
            tracer_provider=tracer_provider,
            meter_provider=meter_provider,
            meter=meter,
            include_messages_counters=False,
        )

    def __call__(self, msg: Any | None) -> BaseMiddleware:
        return PayloadTelemetryMiddleware(
            tracer=self._tracer,
            metrics_container=self._metrics,
            settings_provider_factory=self._settings_provider_factory,
            msg=msg,
        )



from collections.abc import Callable, Coroutine
from typing import Annotated, Any

from fastapi import Depends
from fastapi.security import OAuth2PasswordBearer
from jose.jwt import get_unverified_claims

from fastloom.auth import Claims
from fastloom.auth.protocols import OAuth2Settings
from fastloom.auth.schemas import UserClaims


class OptionalJWTAuth:
    settings: OAuth2Settings
    _oauth2_schema: OAuth2PasswordBearer | None = None

    def __init__(self, settings: OAuth2Settings):
        self.settings = settings
        self._oauth2_schema = OAuth2PasswordBearer(
            str(settings.IAM_TOKEN_URL), auto_error=False
        )

    @classmethod
    def _parse_token(cls, token: str) -> UserClaims:
        return UserClaims.model_validate(get_unverified_claims(token))

    @property
    def get_claims(
        self,
    ) -> Callable[..., Coroutine[Any, Any, UserClaims | None]]:
        async def _inner(
            token: Annotated[str | None, Depends(self._oauth2_schema)],
        ) -> UserClaims | None:
            if token is None:
                return None
            claims = self._parse_token(token)
            Claims.set(claims)
            return claims

        return _inner


class JWTAuth(OptionalJWTAuth):
    def __init__(self, settings: OAuth2Settings):
        super().__init__(settings)
        assert self._oauth2_schema is not None
        self._oauth2_schema.auto_error = True

    @property
    def get_claims(self) -> Callable[..., Coroutine[Any, Any, UserClaims]]:
        async def _inner(
            token: Annotated[str, Depends(self._oauth2_schema)],
        ) -> UserClaims:
            claims = self._parse_token(token)
            Claims.set(claims)
            return claims

        return _inner
